###########################################################################
#                                                                         #
# C O P Y R I G H T   N O T I C E                                         #
#  Copyright (c) 2003-10 by:                                              #
#    * California Institute of Technology                                 #
#                                                                         #
#    All Rights Reserved.                                                 #
#                                                                         #
# Permission is hereby granted, free of charge, to any person             #
# obtaining a copy of this software and associated documentation files    #
# (the "Software"), to deal in the Software without restriction,          #
# including without limitation the rights to use, copy, modify, merge,    #
# publish, distribute, sublicense, and/or sell copies of the Software,    #
# and to permit persons to whom the Software is furnished to do so,       #
# subject to the following conditions:                                    #
#                                                                         #
# The above copyright notice and this permission notice shall be          #
# included in all copies or substantial portions of the Software.         #
#                                                                         #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         #
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF      #
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                   #
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS     #
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN      #
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN       #
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE        #
# SOFTWARE.                                                               #
###########################################################################
#
# motif.py - defines the motif object and its methods in cistematic
from string import upper, lower
from math import log, exp, ceil
from copy import deepcopy
from cistematic.core import complement
from cistematic.stat.score import pearsonCorrelation
import re, os, tempfile

if os.environ.get("CISTEMATIC_ROOT"):
    cisRoot = os.environ.get("CISTEMATIC_ROOT") 
else:
    cisRoot = '/proj/genome'


if os.environ.get("CISTEMATIC_TEMP"):
    cisTemp = os.environ.get("CISTEMATIC_TEMP")
else:
    cisTemp = '/tmp'
tempfile.tempdir = cisTemp

try:
    #import _motif
    import cistematic.core._motif as _motif
    hasMotifExtension = True
except:
    hasMotifExtension = False

matrixRow = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
    
symbolToMatrix = { 'A': [1.0, 0.0, 0.0, 0.0], 'C': [0.0, 1.0, 0.0, 0.0], 'G': [0.0, 0.0, 1.0, 0.0], 
                    'T': [0.0, 0.0, 0.0, 1.0], 'W': [0.5, 0.0, 0.0, 0.5], 'S': [0.0, 0.5, 0.5, 0.0], 
                    'R': [0.5, 0.0, 0.5, 0.0], 'Y': [0.0, 0.5, 0.0, 0.5], 'M': [0.5, 0.5, 0.0, 0.0],
                    'K': [0.0, 0.0, 0.5, 0.5], 'H': [0.33, 0.33, 0.0, 0.34], 'B': [0.0, 0.33, 0.33, 0.34], 
                    'V': [0.33, 0.33, 0.34, 0.0], 'D': [0.33, 0.0, 0.33, 0.34], 'N': [0.25, 0.25, 0.25, 0.25]}
                    

class Motif:
    """ The Motif class is the heart of cistematic. It captures both the consensus, 
        PWM, and Markov(1) versions of the motif, as well as provides methods for scanning 
        sequences for matches.
    """
    motifSeq       = ""
    motifPWM     = []
    reversePWM = []
    motifMarkov1 = []
    revMarkov1 = []
    tagID             = ""
    sequences   = []
    annotations = []
    strictConsensus = ''
    threshold     = 1.0
    info               = ''
    
    def __init__(self, tagID='', theMotif='', thePWM=[], seqs=[], thresh=0.0, info='', motifFile='', seqfile =''):
        """ initialize a Motif object either with (A) a tagID identifier and either of 
            (i) a IUPAC consensus or (ii) a PWM or (iii) example sequences, or (B) from 
            a motif file in Cistematic motif format.
        """        
        fTagID = ''
        if motifFile != '':
            (fTagID, theMotif, thePWM, seqs, thresh, info) = self.readMotif(motifFile)
        
        if len(tagID) > 0:
            self.setTagID(tagID)
        elif len(fTagID) > 0:
            self.setTagID(fTagID)
        else:
            self.setTagID('default')
        
        if theMotif <> '':
            self.setMotif(theMotif)
        
        # PWM overrules the motif
        if len(thePWM) > 1:
            self.setPWM(thePWM)
        
        # a seqfile can be used to create a list of sequences
        if len(seqfile) > 0:
            seqs = self.readSeqFile(seqfile)
            
        # Sequences overrules the PWM and motif
        if len(seqs) > 1:
            self.setSequences(seqs)
            self.setThreshold(len(seqs[0]))
        
        self.setThreshold(float(thresh))
        
        if len(info) > 0:
            self.setInfo(info)
    
    def __len__(self):
        """ returns the length of the motif 
        """
        return len(self.motifPWM)
    
    def readMotif(self, motFile):
        """ read motif in cistmeatic motif format to initialize motif instance.
        """
        fileMotif = ''
        filePWM = []
        fileSeqs = []
        fileThresh = 0.0
        fileInfo = ''
        
        infile = open(motFile, 'r')
        for line in infile:
            if len(line) < 4 or line[0] == '#':
                continue
            
            fields = line.strip().split('\t')
            fieldType = fields[0].lower()
            if fieldType not in ['tagid', 'motif', 'acgt', 'sequence', 'threshold',  'info']:
                print 'could not process line %s' % line
                continue
            if len(fields) < 2:
                continue
            
            if fieldType == 'motif':
                fileMotif = fields[1]
            elif fieldType == 'acgt':
                filePWM.append([float(fields[1]), float(fields[2]), float(fields[3]), float(fields[4])])
            elif fieldType == 'sequence':
                fileSeqs.append(fields[1])
            elif fieldType == 'threshold':
                fileThresh = float(fields[1])
            elif fieldType == 'info':
                fileInfo = fields[1].strip()
            elif fieldType == 'tagid':
                fileTagID = fields[1]
        
        infile.close()
        
        return (fileTagID, fileMotif, filePWM, fileSeqs, fileThresh, fileInfo)
    
    def readSeqFile(self, seqFile):
        """ read sequences from a sequence file.
        """
        fileSeqs = []
        
        infile = open(seqFile, 'r')
        for line in infile:
            fileSeqs.append(line.strip().upper())
        
        infile.close()
        return fileSeqs
        
    def saveMotif(self, motFile):
        """ Save motif in cistematic motif format.
        """
        outfile = open(motFile, 'w')
        
        outfile.write('tagid\t%s\n' % self.tagID)
        outfile.write('info\t%s\n' % self.info.strip())
        outfile.write('threshold\t%s\n' % str(self.threshold))
        outfile.write('motif\t%s\n' % self.buildConsensus())
        for col in self.motifPWM:
            outfile.write('acgt\t%f\t%f\t%f\t%f\n' % (col[0], col[1], col[2], col[3]))
        for seq in self.sequences:
            outfile.write('sequence\t%s\n' % seq)
        
        outfile.close()
    
    def setTagID(self, tag):
        """ set motif identifier.
        """
        self.tagID = tag
    
    def setInfo(self, info):
        """ set motif info string.
        """
        self.info = info
    
    def setThreshold(self, thresh):
        """ set a pre-defined threshold or the highest-possible threshold, otherwise.
        """
        (sForward, sBackward) = self.scoreMotif(self.buildStrictConsensus())
        
        if sForward > sBackward:
            self.threshold = sForward
        else:
            self.threshold = sBackward
        
        if thresh < self.threshold and thresh > 0:
            self.threshold = thresh
    
    def setMotif(self, theMotif):
        """ set the motif PWM using a IUPAC consensus. 
        """
        self.motifSeq = theMotif
        self.motifPWM = self.buildPWM(theMotif)
        self.buildReversePWM()
        self.strictConsensus = self.buildStrictConsensus()
        
    def setSequences(self, theSeqs):
        """ set the founder sequences for the motif and recalculate PWM with them. 
        """
        self.sequences = theSeqs
        
        self.calculatePWMfromSeqs()
        self.calculateMarkov1()
    
    def setPWM(self, thePWM):
        """ set the PWM for the motif and calculate consensus.
        """
        self.motifPWM = thePWM
        self.buildReversePWM()
        self.motifSeq = self.buildConsensus()
        self.strictConsensus = self.buildStrictConsensus()
        
    def appendToPWM(self, col):
        """ add a column to the PWM for the motif 
        """
        self.motifPWM.append(col)
        
    def buildPWM(self, theMotif):
        """ returns the PWM for the provided consensus motif. 
        """
        thePWM = []
        
        for letter in upper(theMotif):
            thePWM.append(symbolToMatrix[letter])
        
        return thePWM
    
    def buildReversePWM(self):
        """ returns the reverse PWM for the motif 
        """
        theRevPWM = []
        motLength = len(self.motifPWM)
        tempPWM = deepcopy(self.motifPWM)
        
        tempPWM.reverse()
        for col in tempPWM:
            theRevPWM.append([col[matrixRow['T']],  col[matrixRow['G']], col[matrixRow['C']], col[matrixRow['A']]])
        
        self.reversePWM = deepcopy(theRevPWM)
        
        return self.reversePWM
        
    def getPWM(self):
        """ returns the PWM for the motif 
        """
        return self.motifPWM
        
    def printPWM(self):
        """ print the PWM and the consensus
        """
        aRow = ''
        cRow = ''
        gRow = ''
        tRow = ''
        
        cons = self.buildConsensus()
        consLine = 'Cons:'
        for NT in cons:
            consLine += '\t'
            consLine += NT
        
        for col in self.motifPWM:
            aRow = aRow + str(round(col[matrixRow['A']],4)) + '\t'
            cRow = cRow + str(round(col[matrixRow['C']],4)) + '\t'
            gRow = gRow + str(round(col[matrixRow['G']],4)) + '\t'
            tRow = tRow + str(round(col[matrixRow['T']],4)) + '\t'
        
        print 'A:\t%s\nC:\t%s\nG:\t%s\nT:\t%s\n' % (aRow, cRow, gRow, tRow)
        print '%s\n' % consLine
    
    def saveLogo(self, outfilename, height=-1, width=-1):
        """ saves a logo version of the motif as outfilename (assumes has .png)
            if the motif is built from sequences.
            will fail if weblogo 2.8.2 package is not installed in correct place.
        """
        logoPath = cisRoot + '/programs/weblogo/seqlogo'
        if outfilename[-4:] in ['.png', '.PNG']:
            outfilename = outfilename[:-4]
        if len(self.sequences) < 1:
            print 'cannot run logo representation without founder sequences'
        else:
            if True:
                seqfilename = tempfile.mktemp() + '.tmp'
                
                seqfile = open(seqfilename,'w')
                for sequence in self.sequences:
                    seqfile.write(sequence + '\n')
                seqfile.flush()
                
                dimensions = ''
                if height > 0:
                    dimensions += '-h %d ' % height
                if width > 0:
                    dimensions +=  '-w %d ' % width
                else:
                    dimensions += '-w %d ' % len(self.motifPWM)
                # /proj/genome/programs/weblogo/seqlogo -f test2-1-cG-1.seqs -F PNG -c -a -Y -n -o cG-1
                cmd = logoPath + ' -f ' + seqfilename + ' -F PNG -c ' + dimensions + '-a -Y -n -M -k 1 -o ' + outfilename
                #print cmd
                contents = os.system(cmd)
                seqfile.close()
                os.remove(seqfilename)
            else:
                print 'failed to make logo: expecting weblogo 2.8.2 package in %s' % logoPath
                print 'also check if ghostscript package is correctly installed.'
    
    def getSymbol(self, col):
        """ helper function for buildConsensus()
        """
        if col[0] > 0.9:
            return 'A'
        if col[1] > 0.9:
            return 'C'
        if col[2] > 0.9:
            return 'G'
        if col[3] > 0.9:
            return 'T'
        
        rVal = ('R', col[0] + col[2])
        yVal = ('Y', col[3] + col[1])
        wVal = ('W', col[0] + col[3])
        sVal = ('S', col[1] + col[2])
        mVal = ('M', col[0] + col[1])
        kVal = ('K', col[3] + col[2])
        
        theVal = rVal
        
        for value in (yVal, wVal, sVal, mVal, kVal):
            if (value[1] > theVal[1]):
                theVal = value
        
        if theVal[1] > 0.9:
            return theVal[0]
        
        bVal = ('B', col[1] + col[2] + col[3])
        dVal = ('D', col[0] + col[2] + col[3])
        hVal = ('H', col[0] + col[1] + col[3])
        vVal = ('V', col[0] + col[1] + col[2])
        
        theVal = bVal
        
        for value in (dVal, hVal, vVal):
            if (value[1] > theVal[1]):
                theVal = value
        
        if theVal[1] > 0.9: 
            return theVal[0]
    
        return 'N'
        
    def buildConsensus(self):
        """ returns the best consensus using the IUPAC alphabet.
        """
        consensus = ''
        for col in self.motifPWM:
            consensus += self.getSymbol(col)
            
        return consensus
    
    def buildStrictConsensus(self):
        """ returns the best consensus using solely nucleotides.
        """
        consensus = ''
        
        for col in self.motifPWM:
            mRow = []
            for nt in ('A', 'C', 'G', 'T'):
                mRow.append((col[matrixRow[nt]], nt))
            mRow.sort()
            consensus += mRow[3][1]
        
        return consensus
        
    def bestConsensusScore(self):
        """ returns the best consensus score possible.
        """
        score = 0.0
        
        for col in self.motifPWM:
            mRow = []
            for nt in ('A', 'C', 'G', 'T'):
                mRow.append((col[matrixRow[nt]], nt))
            mRow.sort()
            score += mRow[3][0]
        
        return score

    def expected(self, length, background=[0.25,0.25, 0.25, 0.25], numMismatches=0):
        """ returns the expected number of matches to the consensus in a sequence of a given length and background. 
        """
        expectedNum = length * self.consensusProbability(background, numMismatches)
        return expectedNum
    
    def consensusProbability(self, background=[0.25,0.25, 0.25, 0.25], numMismatches=0):
        """ returns the probability of the consensus given the background.
        """
        prob = 0
        motifs = []
        if numMismatches> 0:
            seqs = self.seqMismatches(self.buildConsensus().upper(), numMismatches)
            motifs = seqs.split('|')
        else:
            motifs.append(self.buildConsensus())
        for theCons in motifs:
            motProb = 0
            for NT in theCons:
                currentProb = 0.0
                if NT in ('A', 'W', 'R', 'M', 'H', 'V', 'D', 'N'):
                    currentProb += background[matrixRow['A']]
                if NT in ('C', 'S', 'Y', 'M', 'H', 'B', 'V', 'N'):
                    currentProb += background[matrixRow['C']]
                if NT in ('G', 'S', 'R', 'K', 'B', 'V', 'D', 'N'):
                    currentProb += background[matrixRow['G']]
                if NT in ('T', 'W', 'Y', 'K', 'H', 'B', 'D', 'N'):
                    currentProb += background[matrixRow['T']]
                motProb = motProb + log(currentProb)
            prob += exp(motProb)
        return prob
    
    def pwmProbability(self, background):
        """ returns probability of the PWM.
        """
        prob = 1.0
        for row in self.motifPWM:
            currentProb = 0.0
            for NT in ('A', 'C', 'G', 'T'):
                currentProb += row[matrixRow[NT]] * background[matrixRow[NT]]
            prob = prob * currentProb
        
        return prob
    
    def revComp(self):
        """ returns the reverse complement of the consensus of this motif.
        """
        return complement(self.buildConsensus(), len(self.motifPWM))
    
    def numberOfN(self):
        """ returns numbers of effective Ns in motif.
        """
        index = 0
        
        # optimization
        checkSymbol = self.getSymbol
        localPWM = self.motifPWM
        
        for col in localPWM:
            if checkSymbol(col) == 'N':
                index += 1
        
        return index
    
    def buildMismatchSeq(self, rootSeq, tailSeq, numMismatches):
        """ helper function called from seqMismatches().
        """
        finalSeq = ''
        tailLen = len(tailSeq)

        if tailLen < 1 or numMismatches < 1:
            #print 'r = %s \t\t t = %s \t\t n = %i' % (rootSeq, tailSeq, numMismatches)
            return rootSeq + tailSeq

        for pos in range(tailLen - numMismatches + 1):
            newRootSeq = rootSeq
            newRootSeq += tailSeq[:pos]
            newRootSeq += 'N'
            finalSeq += self.buildMismatchSeq(newRootSeq, tailSeq[pos + 1:], numMismatches - 1)
            finalSeq += '|'

        return finalSeq[:-1]

    def seqMismatches(self, seq, numMismatches):
        """ Returns list of sequences that will be used by initializeMismatchRE().
        """
        return self.buildMismatchSeq('', seq, numMismatches)

    
    def probMatchPWM(self, thePWM, NT, colIndex):
        """ returns the probability of seeing that particular nucleotide according to the PSFM.
        """
        currentProb = 0.0
       
        if NT == 'A':
            return thePWM[colIndex][0]
        if NT == 'T':
            return thePWM[colIndex][3]
        if NT == 'C':
            return thePWM[colIndex][1]
        if NT == 'G':
            return thePWM[colIndex][2]
        if NT == 'N':
            return 1.0

        if NT in ('W', 'R', 'M', 'H', 'V', 'D'):
            currentProb += thePWM[colIndex][0]
        if NT in ('S', 'Y', 'M', 'H', 'B', 'V'):
            currentProb += thePWM[colIndex][1]
        if NT in ('S', 'R', 'K', 'B', 'V', 'D'):
            currentProb += thePWM[colIndex][2]
        if NT in ('W', 'Y', 'K', 'H', 'B', 'D'):
            currentProb += thePWM[colIndex][3]
        
        return currentProb
    
    def psfmOdds(self, thePWM, NT, colIndex, background=[0.25, 0.25,0.25,0.25]):
        """ calculates the odds of nucleotide NT coming from position colIndex in thePWM
            as opposed to coming from the background.
        """
        currentProb = 0.0
        
        if NT == 'A':
            return thePWM[colIndex][0]/background[0]
        if NT == 'T':
            return thePWM[colIndex][3]/background[3]
        if NT == 'C':
            return thePWM[colIndex][1]/background[1]
        if NT == 'G':
            return thePWM[colIndex][2]/background[2]
        if NT == 'N':
            return 1.0
        
        backg = 0.0
        if NT in ('W', 'R', 'M', 'H', 'V', 'D'):
            currentProb += thePWM[colIndex][0]
            backg += background[0]
        if NT in ('S', 'Y', 'M', 'H', 'B', 'V'):
            currentProb += thePWM[colIndex][1]
            backg += background[1]
        if NT in ('S', 'R', 'K', 'B', 'V', 'D'):
            currentProb += thePWM[colIndex][2]
            backg += background[2]
        if NT in ('W', 'Y', 'K', 'H', 'B', 'D'):
            currentProb += thePWM[colIndex][3]
            backg += background[3]
        
        return currentProb / backg
    
    def ntMatch(self, motifNT, seqNT):
        """ returns 1 if seqNT matches motifNT.
        """
        if motifNT == seqNT:
            return 1

        if seqNT == 'N' or motifNT == 'N':
            return 1
        
        if motifNT in ('W', 'R', 'M', 'H', 'V', 'D') and seqNT == 'A':
            return 1
        
        if motifNT in ('W', 'Y', 'K', 'H', 'B', 'D') and seqNT == 'T':
            return 1
        
        if motifNT in ('S', 'Y', 'M', 'H', 'B', 'V') and seqNT == 'C':
            return 1
        
        if motifNT in ('S', 'R', 'K', 'B', 'V', 'D') and seqNT == 'G':
            return 1
        
        return 0
    
    def scoreMotif(self, aSeq, diff=1000):
        """ calculates the consensus score using the PSFM
        """
        motLength = len(self.motifPWM)
        if len(aSeq) < motLength:
            return (0.0, 0.0)
        
        matchPWM = self.probMatchPWM
        motPWM = self.motifPWM
        revPWM = self.reversePWM
        
        theSeq = upper(aSeq)
        
        forwardCons = 0.0
        reverseCons = 0.0
        bestCons = 0.0
        for index in range(motLength):
            currentNT = theSeq[index]
            forwardCons += matchPWM(motPWM, currentNT, index)
            reverseCons += matchPWM(revPWM, currentNT, index)
            bestCons += matchPWM(motPWM,self.strictConsensus[index], index)
        if (forwardCons + diff) < bestCons and (reverseCons + diff) < bestCons:
            return (-1, -1)
        return (forwardCons, reverseCons)
         
    def scoreMotifLogOdds(self, aSeq, background=[0.25, 0.25, 0.25, 0.25]):
        """ calculates the log-odds score using the PSFM given the markov(0) background.
        """
        
        motLength = len(self.motifPWM)
        if len(aSeq) < motLength:
            return (0.0, 0.0)
        
        odds = self.psfmOdds
        motPWM = self.motifPWM
        revPWM = self.reversePWM
        
        theSeq = upper(aSeq)
        
        forwardCons = 0.0
        reverseCons = 0.0
        bestCons = 0.0
        for index in range(motLength):
            currentNT = theSeq[index]
            try:
                forwardCons += log(odds(motPWM, currentNT, index, background), 2)
            except:
                forwardCons += log(0.01, 2)
            try:
                reverseCons += log(odds(revPWM, currentNT, index, background), 2)
            except:
                reverseCons += log(0.01, 2)
            bestCons += log(odds(motPWM,self.strictConsensus[index], index, background), 2)
        
        return (forwardCons, reverseCons)
    
    def bestLogOddsScore(self, background=[0.25, 0.25, 0.25, 0.25]):
        """ calculates the best possible log-odds score using the PSFM given the markov(0) background.
        """
        
        motLength = len(self.motifPWM)
        
        odds = self.psfmOdds
        motPWM = self.motifPWM
        
        bestLogOdds = 0.0
        for index in range(motLength):
            bestLogOdds += log(odds(motPWM,self.strictConsensus[index], index, background), 2)
        return bestLogOdds
    
    def locateConsensus(self, aSeq):
        """ returns a list of positions on aSeq that match the consensus exactly.
        """
        cons = self.buildConsensus()
        revComp = self.revComp()
        motLength = len(cons)
        Position = []
        
        if len(aSeq) < motLength:
            return []
        else: 
            theSeq = upper(aSeq)
        
        pos = 0
        seqLength = len(theSeq)
        while pos <= (seqLength - motLength):
            subSeq = theSeq[pos:pos + motLength].strip()
            
            try:
                forwardMot = 1
                for index in range(motLength):
                    if self.ntMatch(cons[index], subSeq[index]) == 0:
                        forwardMot = 0
                        break
                revCompMot = 1
                for index in range(motLength):
                    if self.ntMatch(revComp[index], subSeq[index]) == 0:
                        revCompMot = 0
                        break
            except:
                print 'chocked at pos %d' % pos
                forwardMot = 0
                backwardMot = 1
            
            if forwardMot == 1:
                Position.append((pos, 'F'))
                pos += motLength
            elif revCompMot == 1:
                Position.append((pos, 'R'))
                pos += motLength
            else:
                pos +=1
            
        return Position

    def compareConsensus(self, aSeq):
        """ returns a sequence with nucleotide differences from consensus in lower case.
        """
        cons = self.buildConsensus()
        revComp = self.revComp()
        motLength = len(cons)
                
        if len(aSeq) < motLength:
            raise NameError, 'Sequence too short'
        else: 
            theSeq = upper(aSeq)
        
        forwardMismatch = 0
        backwardMismatch = 0
        forwardSeq = ''
        backwardSeq = ''
        
        for index in range(motLength):
            if self.ntMatch(cons[index], theSeq[index]) == 0:
                forwardMismatch += 1
                forwardSeq += lower(theSeq[index])
            else:
                forwardSeq += theSeq[index]
                    
            if self.ntMatch(revComp[index], theSeq[index]) == 0:
                backwardMismatch += 1
                backwardSeq += lower(theSeq[index])
            else:
                backwardSeq += theSeq[index]
        
        if forwardMismatch <= backwardMismatch:
            return forwardSeq
        else:
            return backwardSeq
    
    def scoreDiffPWM(self, compMotif):
        """ returns a score scaled from 0 (no difference) to 2 (completely different) to  
            quantify the difference between the motif and another motif.
        """
        score = 0.0
        diffPWM = self.getDiffPWM(compMotif)
        for pos in range(len(diffPWM)):
            (adiff, cdiff, gdiff, tdiff) = diffPWM[pos]
            score += abs(adiff) + abs(cdiff) + abs(gdiff) + abs(tdiff)
        
        score /= float(len(diffPWM))
        
        return score
    
    def getDiffPWM(self, compMotif):
        """ subtracts the PWM of compMotif from existing PWM to compare differences. 
            Note that the comparison is only done on the length of the shorter motif.
        """
        diffPWM = []
        compPWM = compMotif.getPWM()
        compMotLen = len(compMotif)
        motLength = len(self.motifPWM)
        compLen = motLength
        
        if compLen < motLength:
            compLen = compMotLength
        
        for pos in range(compLen):
            pwmCol = self.motifPWM[pos]
            pwmColComp = compPWM[pos]
            diffPWM.append([0.0, 0.0, 0.0, 0.0])
            for NT in range(4):
                diffPWM[pos][NT] = pwmCol[NT] - pwmColComp[NT]
        
        return diffPWM
    
    def initializeRE(self):
        """ initializes Regular Expression motif engine.
        """
        global forwardRE
        global backwardRE
        reMAP = {'A': 'A', 'C':'C', 'G': 'G', 'T':'T', 'N':'[ACGT]', 'W':'[AT]', 'S':'[GC]', 'R':'[AG]',
            'Y':'[CT]', 'M':'[AC]', 'K':'[GT]', 'H':'[ACT]', 'B':'[CGT]', 'V':'[ACG]', 'D':'[AGT]', '|':'I'}
        cons = self.buildConsensus().upper()
        revComp = self.revComp().upper()
        
        reCons = ''
        reBackward = ''
        
        for NT in cons:
            reCons += reMAP[NT]
        
        if revComp != cons:
            for NT in revComp:
                reBackward += reMAP[NT]
        else:
            reBackward = 'ZZZZZZZ'
        
        forwardRE = re.compile(reCons, re.IGNORECASE)
        backwardRE = re.compile(reBackward, re.IGNORECASE)
        
    def initializeMismatchRE(self, numMismatches):
        """ initializes Regular Expression motif engine allowing for mismatches.
        """
        global forwardRE
        global backwardRE
        reMAP = {'A': 'A', 'C':'C', 'G': 'G', 'T':'T', 'N':'[ACGT]', 'W':'[AT]', 'S':'[GC]', 'R':'[AG]',
            'Y':'[CT]', 'M':'[AC]', 'K':'[GT]', 'H':'[ACT]', 'B':'[CGT]', 'V':'[ACG]', 'D':'[AGT]', '|':'|'}
        
        cons = self.seqMismatches(self.buildConsensus().upper(), numMismatches)
        revComp = self.seqMismatches(self.revComp().upper(), numMismatches)
        
        reCons = ''
        reBackward = ''
        
        for NT in cons:
            reCons += reMAP[NT]
        
        if self.revComp().upper() != self.buildConsensus().upper():
            for NT in revComp:
                reBackward += reMAP[NT]
        else:
            reBackward = 'ZZZZZZZ'
        
        forwardRE = re.compile(reCons, re.IGNORECASE)
        backwardRE = re.compile(reBackward, re.IGNORECASE)
        
    def locateConsensusRE(self, aSeq):
        """ Returns a list of positions on aSeq that match the consensus exactly. 
            Should be run after either initializeRE() or initializeMismatchRE(numMismatches)
        """
        motLength = len(self.motifPWM)
        Position = []
        Results = []
        
        if len(aSeq) < motLength:
            return []
        
        forwardIter = forwardRE.finditer(aSeq)
        backwardIter = backwardRE.finditer(aSeq)
        
        for match in forwardIter:
            Position.append((match.start(), 'F'))
        
        for match in backwardIter:
            Position.append((match.start(), 'R'))
        
        positionLength = len(Position)
        if positionLength >= 1:
            Position.sort()
            (prevPos, prevSense) = Position[0]
            Results.append((prevPos, prevSense))
            
            for index in range(1, positionLength):
               (pos, sense) = Position[index]
               if pos >= prevPos + motLength:
                    Results.append((pos, sense))
                    (pos, sense) = (prevPos, prevSense)
        return Results
    
    def locateStrictConsensus(self, aSeq, mismatches=0):
        """ returns a list of positions on aSeq that match the strict 
            consensus within some mismatches.
            Only available as a C-extension for greater speed-up for now.
        """
        forwardMer = self.buildStrictConsensus()
        motLength = len(forwardMer)
        revcompMer = complement(forwardMer, motLength)
        
        if hasMotifExtension:
            return  _motif.locateMer(aSeq, forwardMer, revcompMer, mismatches)
        else:
            print 'only supported as part of the C extension for now'
            return []

    def locateMotif(self, aSeq, Threshold=90.0, numberN=0):
        """ returns a list of positions on aSeq that match the PWM within a Threshold, 
            given as a percentage of the optimal consensus score. 
            Will call C-extension for greater speed-up if compiled.
        """
        motLength = len(self.motifPWM)
        seqLength = len(aSeq)
        
        theThreshold = Threshold /  100.0
        if theThreshold < 0.5:
            print 'Threshold less than 50% - will abort locateMotif()'
            return []
        maxScore = self.bestConsensusScore()
        maxDiff = maxScore * (1 - theThreshold)
        
        if seqLength < motLength:
            return []
        else:
            aSeq.strip()
        
        if hasMotifExtension:
            return  _motif.locateMotif(aSeq, self.motifPWM, self.reversePWM, maxScore, maxDiff)
        else:
            theSeq = upper(aSeq)
            Position = []
            pos = 0    
            
            # optimization ?
            scoreMot = self.scoreMotif
            posAppend = Position.append
            
            while pos <= (seqLength - motLength):
                subSeq = theSeq[pos: pos + motLength]
                if subSeq.count('N') > numberN:
                    pos += 1
                    continue
                
                (seqScore, revSeqScore) = scoreMot(subSeq, maxDiff)
                
                if seqScore >= revSeqScore and seqScore > 1.0:
                    posAppend((pos, 'F'))
                elif revSeqScore > 1.0:
                    posAppend((pos, 'R'))
                pos += 1
            
            return Position
           
    def locateMarkov1(self, aSeq, maxFold=5.0):
        """ returns a list of positions on aSeq that match the Markov1 within maxFold.
        """
        motLength = len(self.motifPWM)
        seqLength = len(aSeq)
        
        if maxFold < 1.0:
            print 'maxFold less than 1.0 - will abort locateMarkov1()'
            return []
        maxScore = self.bestMarkov1Score() * maxFold
        
        if seqLength < motLength:
            return []
        else:
            aSeq.strip()
        
        if hasMotifExtension:
            return  _motif.locateMarkov1(aSeq, self.motifMarkov1, self.revMarkov1, maxScore)
            
        theSeq = upper(aSeq)
        Position = []
        pos = 0    
        # optimization ?
        scoreMot = self.scoreMarkov1
        posAppend = Position.append
        while pos <= (seqLength - motLength):
            subSeq = theSeq[pos: pos + motLength]
            if subSeq.count('N') > 0:
                pos += 1
                continue
                
            (seqScore, revSeqScore) = scoreMot(subSeq, maxScore)
                
            if seqScore <= revSeqScore and seqScore < maxScore:
                posAppend((pos, 'F'))
            elif revSeqScore < maxScore:
                posAppend((pos, 'R'))
            pos += 1
            
        return Position
           
    def calculatePWMfromSeqs(self):
        """ calculate the PWM using a set of non-degenerate instances of the motif.
        """
        thePWM = []
        numOccurences = 0
        numSeqs = len(self.sequences)
        
        if numSeqs < 1:
            return
        
        # using length of first sequence as the length of the motif
        length = len(self.sequences[0])
        
        for index in range(length):
            thePWM.append([0.0, 0.0, 0.0, 0.0])
        
        for seq in self.sequences:
            index = 0
            theseq = seq.upper()
            for NT in theseq:
                thePWM[index][matrixRow[NT]] += 1.0
                index += 1
        
        for index in range(length):
            thePWM[index][matrixRow['A']] /= numSeqs
            thePWM[index][matrixRow['C']] /= numSeqs
            thePWM[index][matrixRow['G']] /= numSeqs
            thePWM[index][matrixRow['T']] /= numSeqs
        
        self.motifPWM = thePWM
        self.buildReversePWM()
        self.motifSeq = self.buildConsensus()
        self.strictConsensus = self.buildStrictConsensus()
    
    def printMarkov1(self):
        """ print the Markov1 PSSM of the form previous NT -> current NT.
        """
        aRow = ''
        cRow = ''
        gRow = ''
        tRow = ''
        row = []
        for prior in range(4):
            row.append(['', '', '', ''])
        
        for pos in self.motifMarkov1:
            for prior in range(4):
                for current in range(4):
                    row[prior][current] += str(round(pos[prior][current], 4)) + '\t'
        
        for prior in ['A', 'C', 'G', 'T']:
            for current in ['A', 'C', 'G', 'T']:
                try:
                    print '%s -> %s\t%s\n' % (prior, current, row[matrixRow[prior]][matrixRow[current]])
                except:
                    print 'ERROR: %s %s' % (prior, current)
        print '\n'
        
    def bestMarkov1Score(self):
        """ returns the best markov1 score possible.
        """
        motLength = len(self.motifMarkov1)
        matchMarkov1 = self.probMatchMarkov1
        score = 0.0
        
        for index in range(motLength):
            col = self.motifMarkov1[index]
            mRow = []
            for prior in ['A', 'C', 'G', 'T']:
                for current in ['A', 'C', 'G', 'T']:
                    mRow.append((col[matrixRow[prior]][matrixRow[current]], prior, current))
            mRow.sort()
                
            if index == 0:
                currentProb = matchMarkov1(self.motifMarkov1, 'N', mRow[-1][2], index)
            else:
                currentProb = matchMarkov1(self.motifMarkov1, mRow[-1][1], mRow[-1][2], index)

            if currentProb < 0.0001:
                currentProb = 0.0001
                
            if currentProb > 0.0:
                score -= log(currentProb,2.0)            
        
        return score
    
    def worstMarkov1Score(self):
        """ returns the worst markov1 score possible.
        """
        motLength = len(self.motifMarkov1)
        score = 0.0
        
        for index in range(motLength):
            currentProb = 0.0001
                
            score -= log(currentProb,2.0)            
        
        return score

    def calculateMarkov1(self, pseudoCount=1.0):
        """ calculate the Markov1 PSSM using a set of non-degenerate instances of the motif.
            adds a pseudoCount for unseen combinations.
        """
        self.motifMarkov1 = []
        numSeqs = len(self.sequences) + pseudoCount
        
        if numSeqs < 2:
            return []
        
        # using length of first sequence as the length of the motif
        length = len(self.sequences[0])
        
        for index in range(length):
            self.motifMarkov1.append([[pseudoCount, pseudoCount, pseudoCount, pseudoCount], [pseudoCount, pseudoCount, pseudoCount, pseudoCount], [pseudoCount, pseudoCount, pseudoCount, pseudoCount], [pseudoCount, pseudoCount, pseudoCount, pseudoCount]])
        
        for seq in self.sequences:
            theseq = seq.upper()
            index = 0
            prior = -1
            for pos in theseq:
                if index == 0:
                    for priorNT in range(4):
                        self.motifMarkov1[index][priorNT][matrixRow[pos]] += 0.25
                else:
                    self.motifMarkov1[index][prior][matrixRow[pos]] += 1.0
                prior = matrixRow[pos] 
                index += 1
        
        for index in range(length):
            for prior in range(4):
                for current in range(4):
                    self.motifMarkov1[index][prior][current] /= numSeqs
        
        self.buildReverseMarkov1(pseudoCount)
    
    def buildReverseMarkov1(self, pseudoCount=1.0):
        """ calculate the Reverse Markov1 PSSM using a set of non-degenerate instances of the motif.
        """
        self.revMarkov1 = []
        numSeqs = len(self.sequences) + pseudoCount
        
        if numSeqs < 2:
            return []
        
        # using length of first sequence as the length of the motif
        length = len(self.sequences[0])
        
        for index in range(length):
            self.revMarkov1.append([[pseudoCount, pseudoCount, pseudoCount, pseudoCount], [pseudoCount, pseudoCount, pseudoCount, pseudoCount], [pseudoCount, pseudoCount, pseudoCount, pseudoCount], [pseudoCount, pseudoCount, pseudoCount, pseudoCount]])
        
        for aSeq in self.sequences:
            seq = complement(aSeq.upper(), length)
            index = 0
            prior = -1
            for pos in seq:
                if index == 0:
                    for priorNT in range(4):
                        self.revMarkov1[index][priorNT][matrixRow[pos]] += 0.25
                else:
                    self.revMarkov1[index][prior][matrixRow[pos]] += 1.0
                prior = matrixRow[pos] 
                index += 1
        
        for index in range(length):
            for prior in range(4):
                for current in range(4):
                    self.revMarkov1[index][prior][current] /= numSeqs
    
    def scoreMarkov1(self, aSeq, maxScore=10000000.):
        """calculate the matching score using the Markov1.
            limit search if given a low maxScore
        """
        
        motLength = len(self.motifMarkov1)
        matchMarkov1 = self.probMatchMarkov1
        
        Score = []
        
        if len(aSeq) < motLength:
            return Score
        else:
            theSeq = upper(aSeq)
        
        pos = 0    
        seqProb = 0.0
        revSeqProb = 0.0
            
        subSeq = theSeq[pos: pos + motLength]
            
        previousNT = 'N'
        for index in range(motLength):
            currentNT = subSeq[index]
            currentProb = matchMarkov1(self.motifMarkov1, previousNT, currentNT, index)
            
            if currentProb < 0.0001:
                currentProb = 0.0001
                
            if currentProb > 0.0:
                seqProb -= log(currentProb,2.0)            
            
            try:
                nextNT = subSeq[index + 1]
            except:
                nextNT = 'N'
                
            revCurrentProb = matchMarkov1(self.revMarkov1, previousNT, currentNT, index)
                
            if revCurrentProb < 0.002:
                revCurrentProb = 0.002
                
            if revCurrentProb > 0.0:
                revSeqProb -= log(revCurrentProb, 2.0) 
            
            if seqProb > maxScore and revSeqProb > maxScore:
                return (seqProb, revSeqProb)
            previousNT = currentNT
            
        return (seqProb, revSeqProb)
        
    def probMatchMarkov1(self, theMarkov1, previousNT, NT, colIndex):
        """ returns the likelihood of seeing NT given previousNT at this position of the motif.
        """
        currentProb = 0.0
        
        if NT in ['A', 'C', 'G', 'T']:
            currentNT = matrixRow[NT]
        else: 
            currentNT = 0
        try:
            prevNT = matrixRow[previousNT]
        except:
            for index in range(4):
                currentProb += theMarkov1[colIndex][index][currentNT]
            return currentProb
        
        if NT in [ 'A', 'C', 'G', 'T']:
            return theMarkov1[colIndex][prevNT][currentNT]
        if NT == 'N':
            return 1.0

        if NT in ('W', 'R', 'M', 'H', 'V', 'D'):
            currentProb += theMarkov1[colIndex][prevNT][0]
        if NT in ('S', 'Y', 'M', 'H', 'B', 'V'):
            currentProb += theMarkov1[colIndex][prevNT][1]
        if NT in ('S', 'R', 'K', 'B', 'V', 'D'):
            currentProb += theMarkov1[colIndex][prevNT][2]
        if NT in ('W', 'Y', 'K', 'H', 'B', 'D'):
            currentProb += theMarkov1[colIndex][prevNT][3]
        
        return currentProb
               
    def isSane(self, minLen=7, stretchLen=6):
        """ check for motif sanity, which includes: minimum length, less than half N's in consensus, 
            motifs include more than two nucleotides, no nucleotide or dinucleotide is repeated more 
            than stretchlen. The appropriate exceptions are made for 'GC' dinucleotides.
        """
        stretchLen = int(stretchLen)
        minLen = int(minLen)
        if stretchLen >= minLen:
            stretchLen = minLen - 1
        
        cons = self.buildConsensus()
        motlen = float(len(cons))
        if motlen < minLen:
            return False
        
        anum = cons.count('A')
        gnum = cons.count('G')
        cnum = cons.count('C')
        tnum = cons.count('T')
        nnum = cons.count('N')
        if (nnum >= 0.5 * motlen):
            return False
        if (anum + tnum == motlen) or (anum +  gnum == motlen) or (anum + cnum == motlen) or (gnum + tnum == motlen) or (tnum + cnum == motlen):
            return False
        
        cons = self.buildStrictConsensus()
        if cons.count('A' * stretchLen) or cons.count('G' * stretchLen) or cons.count('C' * stretchLen) or cons.count('T' * stretchLen):
            return False
        if stretchLen / 2 < stretchLen / 2.:
            stretchLen += 1
        if cons.count('AG' * (stretchLen/2)) or cons.count('AC' * (stretchLen/2)) or cons.count('AT' *( stretchLen/2)) or cons.count('CT' *( stretchLen/2)) or cons.count('GT' *( stretchLen/2)):
            return False
        
        return True

def correlateMotifs(actualMotifA, actualMotifB, maxSlide=1):
    """ Compares two motifs using the "pearson correlation coefficient-like" MSS.
        Will slide a motif up to maxSlide bases compared to the other,
        and reports the best score.
    """
    bestScore = 0.0
    
    if len(actualMotifA) < len(actualMotifB):
        motA = actualMotifB
        motB = actualMotifA
    else:
        motA = actualMotifA
        motB = actualMotifB
    
    motApwm = motA.getPWM()
    motBpwm = motB.getPWM()
    motCpwm = motB.buildReversePWM()
    if hasMotifExtension:
        return  _motif.correlateMotifs(motApwm, motBpwm, motCpwm, maxSlide)
    else:
        length = len(motA)
        padLength = length - len(motB)
        Ncol = [symbolToMatrix['N']]
        for slide in range(-1 * maxSlide, maxSlide + padLength + 1):
            pwmA = deepcopy(motApwm)
            pwmB = deepcopy(motBpwm)
            pwmC = deepcopy(motCpwm)
            
            if slide < 0:
                pwmA = Ncol * abs(slide) + pwmA
                pwmB = pwmB + Ncol * (abs(slide) + padLength) 
                pwmC = pwmC + Ncol * (abs(slide) + padLength)
            elif slide > 0 and slide <= maxSlide:
                if padLength > 0:
                    if padLength >= slide:
                        adjustedPadLength = padLength - slide
                        adjustedSlide = 0
                    else:
                        adjustedPadLength = 0
                        adjustedSlide = slide - padLength
                    pwmA = pwmA + Ncol * adjustedSlide
                    pwmB = Ncol * slide + pwmB + Ncol * adjustedPadLength
                    pwmC = Ncol * slide + pwmC + Ncol * adjustedPadLength
                else:
                    pwmA = pwmA + Ncol * slide
                    pwmB = Ncol * slide + pwmB 
                    pwmC = Ncol * slide + pwmC 
            elif slide > maxSlide:
                maxDiff = slide - maxSlide
                pwmA = pwmA + Ncol * maxSlide
                pwmB = Ncol * slide  + pwmB + Ncol * (padLength - maxDiff)
                pwmC = Ncol * slide  + pwmC + Ncol * (padLength - maxDiff)
            else:
                pwmB = pwmB + Ncol * padLength
                pwmC = pwmC + Ncol * padLength            
                
            score1 = 0.0
            score2 = 0.0
            thisLength = len(pwmA)
            for index in range(thisLength):
                score1 += pearsonCorrelation(pwmA[index], pwmB[index])
                score2 += pearsonCorrelation(pwmA[index], pwmC[index])
            
            score1 = score1 / float(thisLength)
            score2 = score2 / float(thisLength)
            if score1 < score2 and score2 > bestScore:
                bestScore = score2
            elif score1 > bestScore:
                bestScore = score1
    
    return bestScore

def MSS(motifA, motifB, maxSlide=1):
    """ Compares two motifs using the motif similarity score (MSS).
        Will slide a motif up to maxSlide bases compared to the other,
        and reports the best score. Wrapper around correlateMotifs()
    """
    return correlateMotifs(motifA, motifB, maxSlide)

def printMSS(motifList, maxSlide=1):
    """ Prints a matrix of MSS comparisons between different motifs
        in motifList.
    """
    for mot1 in motifList:
        print mot1.tagID,
        for mot2 in motifList:
            val = '\t%1.2f' % correlateMotifs(mot1, mot2, maxSlide)
            print val,
        print ''
    
